searchData={"items":[{"type":"module","title":"Makina","doc":"This module provides the macros needed to write Makina models. After macro expansion the resulting\nmodels are property-based testing state machines compatible with the main property-based testing\nlibraries: `PropCheck` or `eqc_ex`.","ref":"Makina.html"},{"type":"module","title":"Writing a model - Makina","doc":"After importing the library and configuring the project you can start writing Makina models. A\nMakina model must be defined inside an Elixir module. To start writing a model you need to import\nthe DSL using the `use` macro:\n\n    defmodule ModelExample do\n      use Makina\n      ...\n    end\n\nOnce the library is imported you can start defining your model. A Makina model is composed by two\nmain parts: a state and some commands that interact with the state. Makina provides the macros\n`state` and `command` for this purpose.","ref":"Makina.html#module-writing-a-model"},{"type":"module","title":"Writing the model state - Makina","doc":"Only one call to `state` is allowed in each model. This macro supports the following syntax:\n\n    state attribute_name: initial_value :: type, ...\n\nwhere:\n\n-   `attribute_name` could be any atom.\n-   `initial_value` could be any Elixir expression.\n-   `type` could be any valid Elixir type. Type definitions are optional, but recommended.","ref":"Makina.html#module-writing-the-model-state"},{"type":"module","title":"Writing the model commands - Makina","doc":"Any number of command definitions is allowed in a Makina model. To write a command, Makina provides\nthe `command` macro. This macro supports the following syntax:\n\n    command command_name(argument :: type, ...) :: return_type do\n      # callbacks\n      ...\n    end\n\nwhere:\n\n-   `command_name` is the name of the command and could be any supported function name.\n-   `type` and `return_type` could be any valid Elixir type. Type definitions are optional, but\n    recommended.\n\nInside the command block the user must write the callbacks that specify the behaviour of the\nmodel. In command definitions callbacks do not need any arguments, these are automatically inserted\nduring macro expansion. Makina commands allow the definition of the callbacks:\n\n-   `pre` is the predicate that should be satisfied to call the command. In this callback all model\n    state attributes are accessible.\n-   `args` is the function that generates the arguments necessary to perform the test call. It can\n    access the state attributes of the model\n-   `valid_args` is a predicate used during shrinking. Checks whether a command can be called after some\n    elements of the trace are removed. It can access the state attributes and the generated arguments.\n-   `valid` this predicate can be used to check if a generated call is valid and therefore is going\n    to modify the state of the model. It can access the state attributes and the generated\n    arguments. The result of valid can be accessed in `next` and `post`.\n-   `call` is the function that calls the system under test. It can only access the arguments of the\n    command.\n-   `next` is a function that generates the next state after calling the command. It can access the\n    arguments, the state and the symbolic result.\n-   `post` is a predicate that must be checked after the command execution. It can access the state, the\n    command arguments and the result of the command execution.\n\nWhen writing a command, callbacks `pre`, `valid_args`, `next` or `post` can be omitted. If omitted,\nthe default implementation of the predicates `pre`, `valid_args` and `post` is `true` and the\ndefault implementation of `next` is the identity function. When the command does not receive any\narguments, the implementation of `args` can be omitted. The remaining callbacks need to be\nobligatorily provided by the user.\n\nThere is a way to provide a default implementation of `call` using the option `implemented_by` in\nthe module configuration:\n\n    defmodule ModelExample do\n      use Makina, implemented_by: Implementation\n      ...\n\n      command command ...\n    end\n\nWith this option enabled, if the module `Implementation` contains a function with the name `command`,\nthe default implementation of `call` will be delegated to the function defined in `Implementation`.","ref":"Makina.html#module-writing-the-model-commands"},{"type":"module","title":"Extending a model - Makina","doc":"Models in Makina can be refined and extended using the built-in extension mechanism provided by\nthe library. To extend a model `A` with some new behaviour in a model `B`, we need to write a new\nmodel `B` which extends `A`. The following option must be used for this purpose:\n\n    defmodule B do\n      use Makina, extends: A\n      ...\n    end\n\nWhen a model `B` extends a model `A` the resulting state is the union of attributes defined in `A` and\n`B`. If a state attribute appears with the same name in `A` and `B`, the attribute in `B` overrides the\ndefinition in `A`.\n\nIn the same way, the resulting commands are the union of commands defined in `A` and `B`. If an\nattribute appears both in `A` and `B`:\n\n-   `pre`, `valid_args`, `valid` and `post` are conjunctive.\n-   `call` is overridden by the implementation in `B`.\n-   `next` and `args` are the merge of the arguments and the updates in `A` and `B`\n    respectively. Updates and arguments in `B` are prioritized.\n\nIn all these callbacks the implementation inherited from the model `A` can be accessed in the model\n`B` using the function `super()`. The `super()` function can optionally provided with the implicit\narguments in each callback:\n\n-   `pre` and `args` can access `super(state)`.\n-   `valid_args` and `valid` can access `super(state, arguments)`.\n-   `call` can access `super(arguments)`.\n-   `next` and `post` can access `super(state, arguments, result)`.","ref":"Makina.html#module-extending-a-model"},{"type":"module","title":"Composing models - Makina","doc":"Models in Makina can also be composed into a new model using the built-in parallel composition\nmechanism provided by the library. To compose the models `A0`, &#x2026; , `AN` into a new model `B`, we need to\nwrite a new model `B` that extends the composition of `A0`, &#x2026; , `AN`. The following syntax is used to\nachieve this:\n\n    defmodule B do\n      use Makina, extends: [A1, ... AN]\n      ...\n    end\n\nThe list provided to `:extends` will generate a composed model where the resulting state is the union\nof all attributes of the models. Commands are also the union of all commands. If the same command\nappears in more than once:\n\n-   `valid_args`, `valid` and `post` are conjunctive.\n-   `next` is the merge of the updates in models `[A1, ... AN]`\n-   `args` makes a random choice between the models whose `pre` callback is true.\n-   `pre` is disjunctive.\n-   `call` makes a random choice between the callbacks that implement this callback.\n\nThe resulting model `B` extends the composition of the models `A0`, &#x2026; `AN`, and also can extend the\nbehaviour of this composition by refining or adding new commands or state attributes.","ref":"Makina.html#module-composing-models"},{"type":"module","title":"Renaming commands - Makina","doc":"When a Makina model is extended there is an option that allows command renaming. The syntax is the\nfollowing:\n\n    defmodule B do\n      use Makina, extends: A, where: [old_name: :new_name]\n      ...\n    end\n\nThis means that the command with the `old_name` will be visible with the name `new_name` in module `B`.","ref":"Makina.html#module-renaming-commands"},{"type":"module","title":"Makina.Module","doc":"","ref":"Makina.Module.html"},{"type":"function","title":"Makina.Module.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Makina.Module.html#child_spec/1"},{"type":"function","title":"Makina.Module.create/3","doc":"","ref":"Makina.Module.html#create/3"},{"type":"function","title":"Makina.Module.ensure_compiled!/1","doc":"","ref":"Makina.Module.html#ensure_compiled!/1"},{"type":"module","title":"Makina.Types","doc":"Provides definitions for types common to all Makina models.","ref":"Makina.Types.html"},{"type":"type","title":"Makina.Types.generator/0","doc":"Represents a PBT generator.","ref":"Makina.Types.html#t:generator/0"},{"type":"type","title":"Makina.Types.generator/1","doc":"","ref":"Makina.Types.html#t:generator/1"},{"type":"type","title":"Makina.Types.symbolic_call/0","doc":"Represents a symbolic call.","ref":"Makina.Types.html#t:symbolic_call/0"},{"type":"type","title":"Makina.Types.symbolic_expr/0","doc":"Represents a symbolic expression.","ref":"Makina.Types.html#t:symbolic_expr/0"},{"type":"type","title":"Makina.Types.symbolic_var/0","doc":"Represents a symbolic var.","ref":"Makina.Types.html#t:symbolic_var/0"},{"type":"type","title":"Makina.Types.unknown/0","doc":"Represents an unknown type.","ref":"Makina.Types.html#t:unknown/0"},{"type":"extras","title":"Makina Library","doc":"# Makina Library\n\nMakina is a testing library that provides a DSL (*domain specific\nlanguage*) for writing PBT (*property-based testing models*) *models*\nfor *stateful* systems.\n\n**Note:** the word *model* is used in PBT to refer to\n*properties* to be checked on the system under test.\n\nMakina has been designed to address principles of *modularity* and\n*model reuse*. Similar to other PBT libraries, Makina uses the Elixir\nmetaprogramming facilities to generate a model in your favourite PBT\nlibrary (like [`PropCheck`](https://github.com/alfert/propcheck.git) or [`eqc_ex`](https://github.com/Quviq/eqc_ex)).","ref":"readme.html"},{"type":"extras","title":"Using Makina - Makina Library","doc":"- Add Makina to your dependencies in `mix.exs`\n```elixir\ndef deps do\n    [\n        {:makina, \"~> 0.3.2\"}\n    ]\nend\n```\n- Configure your preferred PBT compatible library (e.g. `PropCheck`) in your testing environment config file `config/test.exs`:\n```elixir\nconfig :makina, checker: PropCheck\n```","ref":"readme.html#using-makina"},{"type":"extras","title":"Write your model - Makina Library","doc":"To write your tests you will need to write a model. Let's assume you\nneed to test a *counter system*, i.e. a module `Counter` with\nfunctions `put`, `ìnc` and `get`. A *model* of such of a system is:\n\n```elixir\ndefmodule Counter.Model do\n  use Makina, implemented_by: Counter\n\n  state value: 0 :: integer\n\n  invariants greater_than_0: value >= 0\n\n  command inc :: :ok | :error do\n    next value: value + 1\n  end\n\n  command put(n :: integer) :: :ok | :error do\n    args n: integer(0, :inf)\n    next value: n\n  end\n\n  command get() :: integer do\n    post result == value\n  end\nend\n```","ref":"readme.html#write-your-model"},{"type":"extras","title":"Write the test - Makina Library","doc":"The *boilerplate* to write the *property* is like this:\n\n```elixir\ndefmodule MakinaExamplesTest do\n  use PropCheck\n  use ExUnit.Case, async: true\n\n  describe \"Counter\" do\n    property \"Correctness\" do\n      forall cmds <- commands(Counter.Model.Behaviour) do\n        {:ok, _pid} = Counter.start_link(0)\n        r = {_, _, result} = run_commands(model, cmds)\n        Counter.stop()\n\n        (result == :ok)\n        |> when_fail(print_report(r, cmds))\n        |> aggregate(command_names(cmds))\n      end\n    end\n  end\nend\n```","ref":"readme.html#write-the-test"},{"type":"extras","title":"Makina Changelog","doc":"# Makina Changelog","ref":"changelog.html"},{"type":"extras","title":"Legend - Makina Changelog","doc":"- `[+]` Added for new features\n- `[-]` Removed for now removed features\n- `[C]` Changed for changes in existing functionality\n- `[F]` Fixed for any bug fixes\n- `[O]` Obsolete for soon-to-be removed features\n- `[T]` Technical change that doesn't affect the API (eg. refactoring, tooling, etc.)","ref":"changelog.html#legend"},{"type":"extras","title":"Next release - Makina Changelog","doc":"","ref":"changelog.html#next-release"},{"type":"extras","title":"0.3.0 - Makina Changelog","doc":"- `[+]` New Makina constructions:\n  - Command\n    - *abstract* commands.\n    - *valid* (says if the generated call is successful).\n- `[C]` Some changes in the original composition rules and callbacks:\n  - Command\n    - callbacks can now return `nil`.\n  - Composition\n    - extension always refines the base callbacks.\n    - improved parallel composition rules.\n- `[F]` Generated code now doesn't produce Dialyzer warnings.\n- `[T]` Improved internal documentation.\n- `[-]` Tracer removed, it is no longer needed.\n- `[T]` Move tests to this repository.","ref":"changelog.html#0-3-0"},{"type":"extras","title":"0.2.2 - Makina Changelog","doc":"- `[T]` Documentation improved.","ref":"changelog.html#0-2-2"},{"type":"extras","title":"0.2.1 - Makina Changelog","doc":"- `[+]` Initial public release of Makina. Supported constructions:\n  - State\n    - *attribute names*\n    - *initial values*\n    - *types*\n    - *invariants*\n  - Command\n    - *argument names*\n    - *types*\n    - *pre* (precondition)\n    - *arg* (argument generators)\n    - *valid args* (precondition used during shrinking)\n    - *call* (the call to system under test)\n    - *next* (calculates the next state of the model)\n    - *post* (postcondition that compares the system under test result with the model)\n  - Composition\n    - Extension (*inheritance*)\n    - Parallel composition","ref":"changelog.html#0-2-1"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.37.3"}}